def run(mcd, indent = '  '):
  i = indent
  shape_hdr = [
    "/*",
    f"{i}This file was generated by mcd2cpp.py",
    f"{i}It should not be edited by hand.",
    "*/",
    "",
    f"// For MCD version {mcd.version['minecraftVersion']}",
    "",
    "#ifndef SHAPEDATA_HPP",
    "#define SHAPEDATA_HPP",
    "",
    "#include <vector>",
    "",
    "namespace mcd {",
    "",
    "union MCPoint {",
    f"{i}struct {{",
    f"{i*2}double x, y, z;",
    f"{i}}};",
    f"{i}double vec[3];",
    "};",
    "",
    "struct MCBBoxData {",
    f"{i}MCPoint min, max;",
    "};",
    "",
    f"extern const std::vector<MCBBoxData> shapes[{len(mcd.blockCollisionShapes['shapes'])}];",
    "",
    "} // namespace mcd",
    "#endif // SHAPEDATA_HPP",
  ]

  with open(f"shape_data.hpp", "w") as f:
    f.write('\n'.join(shape_hdr))

  shape_cpp = [
    "/*",
    f"{i}This file was generated by mcd2cpp.py",
    f"{i}It should not be edited by hand.",
    "*/",
    "",
    "#include \"shape_data.hpp\"",
    "",
    "namespace mcd {",
    ""
    "const std::vector<MCBBoxData> shapes[] = {",
  ]

  for shape in mcd.blockCollisionShapes['shapes'].values():
    shape_cpp.append(f"{i}{{{'' if shape else '},'}")
    if shape:
      shape_cpp.append(',\n'.join(
          f"{i*2}{{\n"
          f"{i*3}.min = {{{box[0]}, {box[1]}, {box[2]}}},\n"
          f"{i*3}.max = {{{box[3]}, {box[4]}, {box[5]}}},\n"
          f"{i*2}}}"
      for box in shape))
      shape_cpp.append(f"{i}}},")
  shape_cpp[-1] = shape_cpp[-1][:-1]

  shape_cpp.extend((
    "};",
    "",
    "} // namespace mcd",
  ))

  with open(f"shape_data.cpp", "w") as f:
    f.write('\n'.join(shape_cpp))



# Copied wholesale from SpockBot, thanks Gjum
import copy


def get_settings(defaults, settings):
  return dict(copy.deepcopy(defaults), **settings)


def pl_announce(*args):
  def inner(cl):
    cl.pl_announce = args
    return cl

  return inner


class PluginBase:
  """A base class for cleaner plugin code.
  Extending from PluginBase allows you to declare any requirements, default
  settings, and event listeners in a declarative way. Define the appropriate
  attributes on your subclass and enjoy cleaner code.
  """
  requires = ()
  defaults = {}
  events = {}

  def __init__(self, ploader, settings):
    # Load the plugin's settings.
    self.settings = get_settings(self.defaults, settings)

    # Load all the plugin's dependencies.
    if isinstance(self.requires, str):
      setattr(self, self.requires.lower(), ploader.require(self.requires))
    else:
      for requirement in self.requires:
        setattr(self, requirement.lower(), ploader.require(requirement))

    # Setup the plugin's event handlers.
    if self.events:
      ev = ploader.require('Event')
      for event in self.events.keys():
        if hasattr(self, self.events[event]):
          ev.register_callback(event, getattr(self, self.events[event]))
        else:
          raise AttributeError(f"{self.__class__.__name__} object has no "
                               f"attribute '{self.events[event]}'")


# Most C Plugins are used only for dependency resolution, they don't stick
# around to own their core. Therefore we need to release ownership at
# the Python level and hand it over to the plugin loader. This plugin does that
# generically
class CPluginBase(PluginBase):
  core = None

  def __init__(self, ploader, settings):
    self.core(ploader, True).thisown = 0



from hashlib import sha1
from urllib import request, error
import json

from rikerbot import logger
from rikerbot.proto.yggdrasil import Yggdrasil
from rikerbot.PluginBase import PluginBase, pl_announce


class AuthCore:
  def __init__(self, event, online_mode, auth_timeout):
    self.event = event
    self.online_mode = online_mode
    self.auth_timeout = auth_timeout
    self.ygg = Yggdrasil()
    self._username = None
    self.login_success = event.register_event('auth_login_success')
    self.login_error = event.register_event('auth_login_error')

  def set_username(self, username):
    self.ygg.username = username

  username = property(lambda self: self._username, set_username)

  def set_password(self, password):
    self.ygg.password = password

  password = property(lambda self: bool(self.ygg.password), set_password)

  def set_client_token(self, token):
    self.ygg.client_token = token

  client_token = property(lambda self: self.ygg.client_token, set_client_token)

  def set_access_token(self, token):
    self.ygg.access_token = token

  access_token = property(lambda self: self.ygg.access_token, set_access_token)

  def set_access_token(self, token):
    self.ygg.access_token = token

  access_token = property(lambda self: self.ygg.access_token, set_access_token)

  selected_profile = property(lambda self: self.ygg.selected_profile)

  def login(self):
    if not self.online_mode:
      self._username = self.ygg.username
      return True
    if self.ygg.login():
      self._username = self.ygg.selected_profile['name']
      logger.info(f"Successful Login, username is: {self._username}")
      self.event.emit(self.login_success)
      return True
    self.event.emit(self.login_error)


@pl_announce('Auth')
class AuthPlugin(PluginBase):
  requires = ('Event', 'IO')
  events = {
      'ClientboundEncryptionBegin': 'handle_session_auth',
  }
  defaults = {
      'online_mode': True,
      'auth_timeout': 3,  # No idea how long this should be, 3s seems good
      'auth_quit': True,
      'sess_quit': True,
  }

  def __init__(self, ploader, settings):
    super().__init__(ploader, settings)
    self.auth_timeout = self.settings['auth_timeout']
    self.core = AuthCore(self.event, self.auth_timeout,
                         self.settings['online_mode'])
    ploader.provide('Auth', self.core)
    self.session_auth = self.event.register_event("auth_session_success")

  def handle_session_auth(self, event_id, packet):
    # Server ID is blank for Notchian servers, but some custom servers sitll
    # use it
    sev_id = packet.serverId.encode('ascii')
    digest = sha1(sev_id + self.io.shared_secret + packet.publicKey).digest()
    data = json.dumps({
        'accessToken':
        self.core.ygg.access_token,
        'selectedProfile':
        self.core.ygg.selected_profile['id'],
        'serverId':
        format(int.from_bytes(digest, 'big', signed=True), 'x')
    }).encode('utf-8')
    url = 'https://sessionserver.mojang.com/session/minecraft/join'
    req = request.Request(url, data, {'Content-Type': 'application/json'})
    try:
      rep = request.urlopen(req, timeout=self.auth_timeout)
      rep = rep.read().decode('utf-8')
    except error.URLError:
      rep = "Couldn't connect to sessionserver.mojang.com"
    if rep:
      logger.warning(rep)
    else:
      logger.info("Successful Session Auth")
      self.event.emit(self.session_auth, packet,
                      "mcd::ClientboundEncryptionBegin *")



from rikerbot.PluginBase import CPluginBase, pl_announce
from .CStatusCore import StatusCore


@pl_announce('Status')
class StatusPlugin(CPluginBase):
  requires = ("Event", "IO")
  core = StatusCore



# if it doesn't come bursting out of you
# in spite of everything,
# don't do it.
# unless it comes unasked out of your
# heart and your mind and your mouth
# and your gut,
# don't do it.
# if you have to sit for hours
# staring at your computer screen
# or hunched over your
# typewriter
# searching for words,
# don't do it
#   - Charles Bukowski, so you want to be a writer?

import rikerbot.proto
import rikerbot.CLogger as logger

from rikerbot.proto import mc_position, mc_uuid

from .plugins.IOPlugin import IOPlugin
from .plugins.EventPlugin import EventPlugin
from .plugins.ExecPlugin import ExecPlugin
from .plugins.StartPlugin import StartPlugin
from .plugins.AuthPlugin import AuthPlugin
from .plugins.KeepAlive import KeepAlivePlugin
from .plugins.WorldPlugin import WorldPlugin
from .plugins.StatusPlugin import StatusPlugin

from .DependencySolver import solve_dependencies
from .PluginBase import PluginBase, pl_announce
from .PluginLoader import PluginLoader, make_PluginLoader, delete_PluginLoader

default_plugins = [
    ('start', StartPlugin),
    ('io', IOPlugin),
    ('event', EventPlugin),
    ('auth', AuthPlugin),
    ('keepalive', KeepAlivePlugin),
    ('world', WorldPlugin),
    ('status', StatusPlugin),
    ('exec', ExecPlugin),
]

from .SimpleClient import SimpleClient



# What are the roots that clutch, what branches grow
# Out of this stony rubbish? Son of man,
# You cannot say, or guess, for you know only
# A heap of broken images, where the sun beats,
# And the dead tree gives no shelter, the cricket no relief
#   - T.S. Eliot, The Waste Land

indent = "  "

mcd_typemap = {}
def mc_data_name(typename):
  def inner(cls):
      mcd_typemap[typename] = cls
      return cls
  return inner

# MCD/Protodef has two elements of note, "fields" and "types"
# "Fields" are JSON objects with the following members:
#   * "name" (optional): Name of the field
#   * "anonymous" (optional): Only present if Name is absent, always "True"
#   * "type": A MCD/Protodef "type"
#
# "Types" are either strings or 2-element JSON arrays
#   * String "types" are a string of the type name
#   * JSON Array "types" have the following elements:
#     * 0: String of the type name, same as String "types"
#     * 1: Type data, a JSON object who's members are type-dependent
#
# Of Note: The only MCD "type" that can contain other MCD "fields" is the
# "container" type. Even though MCD "switches" have a type data member called
# "fields" these are not MCD "fields", they are actually MCD "types".
#
# The following extract functions normalize this format

# Param: MCD "type"
# Returns: Field type, Field data
def extract_type(type_field):
  if(isinstance(type_field, str)):
    return type_field, []
  return type_field[0], type_field[1]

# Param: MCD "field"
# Returns: Field name, Field type, Field data
def extract_field(packet_field):
  name = packet_field.get('name', '')
  field_type, field_data = extract_type(packet_field['type'])
  return name, field_type, field_data

class generic_type:
  typename = ''
  postfix = ''

  def __init__(self, name, parent, type_data = None, use_compare = False):
    self.name = name
    self.old_name = None
    self.parent = parent
    self.switched = False

    # This is needed for switches because switches suck ass
    # More descriptively, switches need to be able to change the names of
    # fields when they're constructed in order to avoid name collisions. But
    # when merging later sister fields they need to be able to identify
    # duplicate fields that didn't undergo that transformation.
    #
    # To accomplish this, we repurpose the "name" field strictly as a display
    # name, and do comparisons with the "compare_name".
    self.compare_name = name
    self.use_compare = use_compare

  def declaration(self):
    return (f"{self.typename} {self.name};",)

  def initialization(self, val):
    return (f"{self.typename} {self.name} = {val};",)

  def encoder(self):
    return (f"enc_{self.postfix}(dest, {self.name});",)

  def decoder(self):
    if self.name:
      return (f"{self.name} = dec_{self.postfix}(src);",)
    # Special case for when we need to decode a variable as a parameter
    else:
      return f"dec_{self.postfix}(src)"

  def dec_initialized(self):
    return (f"{self.typename} {self.name} = dec_{self.postfix}(src);",)

  # This comes up enough to write some dedicated functions for it
  # Conglomerate types take one of two approaches to fundamental types:
  # * Set the field name _every_ time prior to decl/enc/dec
  # * Only change it when needed, and then reset it at the end of the op
  # These support the second workflow
  def temp_name(self, temp):
    if self.old_name is None:
      self.old_name = self.name
    self.name = temp

  def reset_name(self):
    if self.old_name is not None:
      self.name = self.old_name

  def __eq__(self, val):
    if not isinstance(val, type(self)):
      return False
    if hasattr(self, 'typename') and hasattr(val, 'typename'):
      if self.typename != val.typename:
        return False
    if self.use_compare:
      return self.compare_name == val.compare_name
    return self.name == val.name

  def __str__(self):
    return f"{self.typename} {self.name}"

# A "simple" type is one that doesn't need to typedef an additional type in
# order to declare its variable. The only "complex" types are bitfields and
# containers, everything else is considered "simple"
class simple_type(generic_type):
  pass

class numeric_type(simple_type):
    size = 0

# These exist because MCD switches use them. I hate MCD switches
@mc_data_name('void')
class void_type(numeric_type):
  typename = 'void'
  def declaration(self):
    return (f"// '{self.name}' is a void type",)

  def encoder(self):
    return (f"// '{self.name}' is a void type",)

  def decoder(self):
    return (f"// '{self.name}' is a void type",)

@mc_data_name('u8')
class num_u8(numeric_type):
  size = 1
  typename = 'std::uint8_t'
  postfix = 'byte'

@mc_data_name('i8')
class num_i8(num_u8):
  typename = 'std::int8_t'

@mc_data_name('bool')
class num_bool(num_u8):
  pass

@mc_data_name('u16')
class num_u16(numeric_type):
  size = 2
  typename = 'std::uint16_t'
  postfix = 'be16'

@mc_data_name('i16')
class num_i16(num_u16):
  typename = 'std::int16_t'

@mc_data_name('u32')
class num_u32(numeric_type):
  size = 4
  typename = 'std::uint32_t'
  postfix = 'be32'

@mc_data_name('i32')
class num_i32(num_u32):
  typename = 'std::int32_t'

@mc_data_name('u64')
class num_u64(numeric_type):
  size = 8
  typename = 'std::uint64_t'
  postfix = 'be64'

@mc_data_name('i64')
class num_i64(num_u64):
  typename = 'std::int64_t'

@mc_data_name('f32')
class num_float(num_u32):
  typename = 'float'
  postfix = 'bef32'

@mc_data_name('f64')
class num_double(num_u64):
  typename = 'double'
  postfix = 'bef64'

# Positions and UUIDs are broadly similar to numeric types
# A position is technically a bitfield but we hide that behind a utility func
@mc_data_name('position')
class num_position(num_u64):
  typename = 'mc_position'
  postfix = 'position'

@mc_data_name('UUID')
class num_uuid(numeric_type):
  size = 16
  typename = 'mc_uuid'
  postfix = 'uuid'

@mc_data_name('varint')
class mc_varint(numeric_type):
  # typename = 'std::int32_t'
  # All varints are varlongs until this gets fixed
  # https://github.com/PrismarineJS/minecraft-data/issues/119
  typename = 'std::int64_t'
  postfix = 'varint'

@mc_data_name('varlong')
class mc_varlong(numeric_type):
  typename = 'std::int64_t'
  # Decoding varlongs is the same as decoding varints
  postfix = 'varint'

@mc_data_name('string')
class mc_string(simple_type):
  typename = 'std::string'
  postfix = 'string'

@mc_data_name('buffer')
class mc_buffer(simple_type):
  typename = 'std::vector<char>'
  postfix = 'buffer'

  def __init__(self, name, parent, type_data, use_compare = False):
    super().__init__(name, parent, type_data, use_compare)
    self.count = mcd_typemap[type_data['countType']].postfix

  def encoder(self):
    return (
      f"enc_{self.count}(dest, {self.name}.size());",
      f"enc_buffer(dest, {self.name});",
    )

  def decoder(self):
    return (f"{self.name} = dec_buffer(src, dec_{self.count}(src));",)

@mc_data_name('restBuffer')
class mc_restbuffer(simple_type):
  typename = 'std::vector<char>'
  postfix = 'buffer'

  def encoder(self):
    return (f"dest.write({self.name}.data(), {self.name}.size());",)

  def decoder(self):
    return (
      f"{self.name} = std::vector<char>(std::istreambuf_iterator<char>(src),",
      indent*2 + f"std::istreambuf_iterator<char>());"
    )

@mc_data_name('nbt')
class mc_nbt(simple_type):
  typename = 'nbt::TagCompound'

  def encoder(self):
    return (f"{self.name}.encode_full(dest);",)

  def decoder(self):
    return (f"{self.name}.decode_full(src);",)

@mc_data_name('optionalNbt')
class mc_optnbt(simple_type):
  typename = 'std::optional<nbt::TagCompound>'

  def encoder(self):
    return (
      f"if({self.name})",
      f"{indent}{self.name}->encode_full(dest);",
      f"else",
      f"{indent}enc_byte(dest, nbt::TAG_END);"
    )

  def decoder(self):
    return (
      "if(dec_byte(src) == nbt::TAG_COMPOUND)",
      f"{indent}{self.name}.emplace(src, nbt::read_string(src));",
    )

class self_serializing_type(simple_type):
  def encoder(self):
    return (f"{self.name}.encode(dest);",)

  def decoder(self):
    return (f"{self.name}.decode(src);",)

@mc_data_name('slot')
class mc_slot(self_serializing_type):
  typename = 'MCSlot'

@mc_data_name('minecraft_smelting_format')
class mc_smelting(self_serializing_type):
  typename = 'MCSmelting'

@mc_data_name('entityMetadata')
class mc_metadata(self_serializing_type):
  typename = 'MCEntityMetadata'

# This is not how topBitSetTerminatedArray works, but the real solution is hard
# and this solution is easy. As long as this type is only found in the Entity
# Equipment packet we're going to stick with this solution
@mc_data_name('topBitSetTerminatedArray')
class mc_entity_equipment(self_serializing_type):
  typename = 'MCEntityEquipment'

@mc_data_name('particleData')
class mc_particle(self_serializing_type):
  typename = 'MCParticle'
  def __init__(self, name, parent, type_data, use_compare = False):
    super().__init__(name, parent, type_data, use_compare)
    self.id_field = type_data['compareTo']

  def decoder(self):
    return (f"{self.name}.decode(src, "
        f"static_cast<particle_type>({self.id_field}));",)

class vector_type(simple_type):
  count = mc_varint
  def __init__(self, name, parent, type_data, use_compare = False):
    super().__init__(name, parent, type_data, use_compare)
    self.depth = 0
    p = parent
    while not isinstance(p, packet):
      self.depth += 1
      p = p.parent

  def encoder(self):
    return (
      f"enc_{self.count.postfix}(dest, {self.name}.size());",
      f"for(auto &el{self.depth} : {self.name})",
      f"{indent}el{self.depth}.encode(dest);"
    )

  def decoder(self):
    return (
      f"{self.name}.resize(dec_{self.count.postfix}(src));",
      f"for(auto &el{self.depth} : {self.name})",
      f"{indent}el{self.depth}.decode(src);"
    )

@mc_data_name('ingredient')
class mc_ingredient(vector_type):
  typename = 'std::vector<MCSlot>'

@mc_data_name('tags')
class mc_tags(vector_type):
  typename = 'std::vector<MCTag>'

@mc_data_name('option')
class mc_option(simple_type):
  def __init__(self, name, parent, type_data, use_compare = False):
    super().__init__(name, parent, type_data, use_compare)
    f_type, f_data = extract_type(type_data)
    self.field = mcd_typemap[f_type](f"{name}", self, f_data)

    if isinstance(self.field, simple_type):
      self.typename = f"std::optional<{self.field.typename}>"
    else:
      self.typename = f"std::optional<{self.name}_type>"

  def declaration(self):
    if isinstance(self.field, simple_type):
      return super().declaration()
    self.field.name = f"{self.name}_type"
    return [
      *self.field.typedef(),
      f"{self.typename} {self.name};"
    ]

  def encoder(self):
    self.field.name = f"{self.name}.value()"
    return [
      f"enc_byte(dest, {self.name}.has_value());",
      f"if({self.name}) {{",
      *(indent + line for line in self.field.encoder()),
      "}"
    ]

  def decoder(self):
    ret = [f"if(dec_byte(src)) {{"]
    if isinstance(self.field, numeric_type) or type(self.field) in (mc_string,
        mc_buffer, mc_restbuffer):
      self.field.name = self.name
    elif isinstance(self.field, vector_type) or isinstance(self.field,
        complex_type):
      ret.append(f"{indent}{self.name}.emplace();")
      self.field.name = f"{self.name}.value()"
    else:
      self.field.name = f"{self.name}.emplace()"
    ret.extend(indent + line for line in self.field.decoder())
    ret.append('}')
    return ret

class complex_type(generic_type):
  def declaration(self):
    if self.name:
      return [
        "struct {",
        *(indent + l for f in self.fields for l in f.declaration()),
        f"}} {self.name};"
      ]
    return [l for f in self.fields for l in f.declaration()]

  def typedef(self):
    ret = self.declaration()
    ret[0] = f"typedef {ret[0]}"
    return ret

def get_storage(numbits):
  if numbits <= 8:
    return 8
  elif numbits <= 16:
    return 16
  elif numbits <= 32:
    return 32
  else:
    return 64

@mc_data_name('bitfield')
class mc_bitfield(complex_type):
  def __init__(self, name, parent, type_data, use_compare = False):
    super().__init__(name, parent, type_data, use_compare)
    lookup_unsigned = {
        8: num_u8,
        16: num_u16,
        32: num_u32,
        64: num_u64
    }
    lookup_signed = {
        8: num_i8,
        16: num_i16,
        32: num_i32,
        64: num_i64
    }
    self.fields = []
    self.extra_data = []
    self.field_sizes = {}

    total = 0
    for idx, field in enumerate(type_data):
      total += field['size']
      if field['name'] in ('_unused', 'unused'):
        continue
      self.field_sizes[field['name']] = field['size']
      shift = 0
      for temp in type_data[idx + 1:]:
        shift += temp['size']
      self.extra_data.append(((1<<field['size'])-1, shift, field['size'],
          field['signed']))
      numbits = get_storage(field['size'])
      if field['signed']:
        self.fields.append(lookup_signed[numbits](field['name'], self))
      else:
        self.fields.append(lookup_unsigned[numbits](field['name'], self))

    self.storage = lookup_unsigned[total](f"{name}_", self)
    self.size = total//8

  def encoder(self):
    ret = [*self.storage.initialization("0")]
    name = f"{self.name}." if self.name else ""
    for idx, field in enumerate(self.fields):
      mask, shift, size, signed = self.extra_data[idx]
      shift_str = f"<<{shift}" if shift else ""
      ret.append(f"{self.storage.name} |= "
          f"(static_cast<{self.storage.typename}>({name}"
          f"{field.name})&{mask}){shift_str};")
    ret.extend(self.storage.encoder())
    return ret

  def decoder(self):
    ret = [*self.storage.dec_initialized()]
    name = f"{self.name}." if self.name else ""
    for idx, field in enumerate(self.fields):
      mask, shift, size, signed = self.extra_data[idx]
      shift_str = f">>{shift}" if shift else ""
      ret.append(f"{name}{field.name} = "
          f"({self.storage.name}{shift_str})&{mask};")
      if signed:
        ret.extend((
          f"if({name}{field.name} & (1UL << {size - 1}))",
          f"{indent}{name}{field.name} -= 1UL << {size};"
        ))
    return ret

# Whatever you think an MCD "switch" is you're probably wrong
@mc_data_name('switch')
class mc_switch(simple_type):
  def __init__(self, name, parent, type_data, use_compare = False):
    super().__init__(name, parent, type_data, use_compare)
    self.compareTo = type_data['compareTo']
    # Unions are a sum type that differs based on conditions. They are a true
    # "switch". All others are here because MCData is bad and I hate it.
    self.is_union = False

    # A non-union switch consists of one or more consistent types, some or all
    # of which are decoded under different conditions. They act like unions
    # except they leave behind Null switches if they're multi-element.
    #
    # Null switches are fields that got merged into a sister switch, they're
    # carcasses that don't do anything. They're found when MCD encodes a
    # multi-element switch. All the elements except the "lead" element become
    # null switches.
    self.null_switch = False
    # Inverses are default present _except_ for certain conditions, and they're
    # bugged on tons of corner cases that thankfully aren't present in MCD
    # right now.
    self.is_inverse = False
    self.is_str_switch = False

    self.lead_sister = None
    self.fields = []
    self.field_dict = {}

    # Find out if all possible values are the same type, in which case we're
    # not a union. And if all the types are void we're an inverse.
    values = type_data['fields'].values()
    if(all(map(lambda x : x == 'void', values))):
      self.is_inverse = True
      f_type, f_data = extract_type(type_data['default'])
      self.fields.append(mcd_typemap[f_type](name, self, f_data))
    else:
      values = list(filter(lambda x : x != 'void', values))
      self.is_union = not all(x == values[0] for x in values)

    # In the case we're not a union, we need to check for sister switches
    if not self.is_union:
      for field in parent.fields:
        if isinstance(field, mc_switch) and field.compareTo == self.compareTo:
          self.null_switch = True
          self.lead_sister = field.lead_sister if field.null_switch else field
          self.lead_sister.merge(name, type_data['fields'], self.is_inverse,
              type_data.get('default', None))
          return

    self.process_fields(self.name, type_data['fields'])

  def declaration(self):
    if self.null_switch:
      return []
    return [l for f in self.fields for l in f.declaration()]

  def code_fields(self, ret, fields, encode = True):
    for field in fields:
      if hasattr(self.parent, 'name') and self.parent.name:
        field.temp_name(f"{self.parent.name}.{field.name}")
      if encode:
        ret.extend(f"{indent}{l}" for l in field.encoder())
      else:
        ret.extend(f"{indent}{l}" for l in field.decoder())
      field.reset_name()
    return ret

  def inverse(self, comp, encode = True):
    if len(self.field_dict.items()) == 1:
      case, _ = next(iter(self.field_dict.items()))
      ret = [f"if({comp} != {case}) {{"]
    else:
      return ('// Multi-Condition Inverse Not Yet Implemented',)
    self.code_fields(ret, self.fields, encode)
    ret.append("}")
    return ret

  # Special case for single condition switches, which are just optionals
  # mascarading as switches
  def optional(self, comp, encode = True):
    case, fields = next(iter(self.field_dict.items()))
    ret = []
    if case == "true":
      ret.append(f"if({comp}) {{")
    elif case == "false":
      ret.append(f"if(!{comp}) {{")
    elif case.isdigit():
      ret.append(f"if({comp} == {case}) {{")
    else:
      ret.append(f"if(!{comp}.compare({case})) {{")
    self.code_fields(ret, fields, encode)
    ret.append("}")
    return ret

  def str_switch(self, comp, encode = True):
    items = list(self.field_dict.items())
    case, fields = items[0]
    ret = [f"if(!{comp}.compare({case})) {{"]
    self.code_fields(ret, fields, encode)
    for case, fields in items[1:]:
      ret.append(f"}} else if(!{comp}.compare({case})) {{")
      self.code_fields(ret, fields, encode)
    ret.append("}")
    return ret

  def union_multi(self, comp, encode = True):
    if len(self.field_dict.items()) == 1:
      return self.optional(comp, encode)
    ret = [f"switch({comp}) {{"]
    for case, fields in self.field_dict.items():
      ret.append(f"{indent}case {case}:")
      tmp = []
      self.code_fields(tmp, fields, encode)
      ret.extend(indent + l for l in tmp)
      ret.append(f"{indent*2}break;")
    ret.append("}")
    return ret

  # compareTo is a mystical format, its powers unmatched. We don't try to parse
  # it. Instead, we just count the number of hierarchy levels it ascends with
  # ".." and move up the container hierarchy using that. If we hit the packet,
  # we abandon ship and assume the path is absolute.
  def get_compare(self):
    comp = self.compareTo.replace('../','').replace('..','').replace('/','.')
    p = self.parent
    for i in range(self.compareTo.count('..')):
      while not (isinstance(p, complex_type) or isinstance(p, packet)):
        p = p.parent
      if isinstance(p, packet):
        break;
      else:
        p = p.parent
    while not (isinstance(p, complex_type) or isinstance(p, packet)):
      p = p.parent
    if not isinstance(p, packet):
      comp = f"{p.name}.{comp}"
    return comp

  def encoder(self):
    comp = self.get_compare()
    if self.null_switch:
      return []
    if self.is_inverse:
      return self.inverse(comp)
    if self.is_str_switch:
      return self.str_switch(comp)
    return self.union_multi(comp)

  def decoder(self):
    comp = self.get_compare()
    if self.null_switch:
      return []
    if self.is_inverse:
      return self.inverse(comp, False)
    if self.is_str_switch:
      return self.str_switch(comp, False)
    return self.union_multi(comp, False)

  def process_fields(self, name, fields):
    for key, field_info in fields.items():
      if not key.isdigit() and key not in ("true", "false"):
        key = f'"{key}"'
        self.is_str_switch = True
      f_type, f_data = extract_type(field_info)
      field = mcd_typemap[f_type](name, self, f_data, True)
      if field not in self.fields:
        if not isinstance(field, void_type):
          self.fields.append(field)
      if not self.is_inverse and isinstance(field, void_type):
        continue
      if key in self.field_dict:
        self.field_dict[key].append(field)
      else:
        self.field_dict[key] = [field]

      # Look for name collisions and fix them. This is unnecessarily convoluted
      # and ripe for refactoring.
      has_dupes = []
      for field in self.fields:
        if len(list(filter(lambda x: x.compare_name == field.compare_name,
            self.fields))) > 1:
          has_dupes.append(field)
      # If we're an anonymous switch don't bother, we'll break container fields
      if has_dupes and name:
        for key, fields in self.field_dict.items():
          # Is this fine? Is it guaranteed there will be (at least) one and
          # only one dupe for each key? I think so?
          same = next(x for x in fields if x in has_dupes)
          if self.is_str_switch:
            tmp = key.replace('"', '').replace(':', '_')
            same.name = tmp
            # As a weird consequence of the merging strategy, we can end up in
            # a situation with two tags that compare the same but have
            # different names. These tags have the same type, but because
            # string switch fields have "convenient" names they have different
            # names. Here we restore these "lost" tags to the field list
            if all(map(lambda x: x.name != same.name, self.fields)):
              self.fields.append(same)
          else:
            same.name = f"{name}_{key}"

      if not self.is_str_switch:
        self.field_dict = {key: self.field_dict[key] for key in
            sorted(self.field_dict)}

  def merge(self, sis_name, sis_fields, is_inverse, default):
    if is_inverse:
      f_type, f_data = extract_type(default)
      self.fields.append(mcd_typemap[f_type](sis_name, self, f_data))
    self.process_fields(sis_name, sis_fields)

  def __eq__(self, value):
    if not super().__eq__(value) or len(self.fields) != len(value.fields):
        return False
    return all([i == j for i, j in zip(self.fields, value.fields)])


# Arrays come in three flavors. Fixed length, length prefixed, and foreign
# field
@mc_data_name('array')
class mc_array(simple_type):
  def __init__(self, name, parent, type_data, use_compare = False):
    super().__init__(name, parent, type_data, use_compare)
    f_type, f_data = extract_type(type_data['type'])
    self.field = mcd_typemap[f_type]('', self, f_data)
    self.depth = 0
    p = parent
    while not isinstance(p, packet):
      self.depth += 1
      p = p.parent
    self.packet = p

    self.is_fixed = False
    self.is_prefixed = False
    self.is_foreign = False
    self.count = None
    if "countType" in type_data:
      self.is_prefixed = True
      self.count = mcd_typemap[type_data['countType']]('', self, [])
    elif isinstance(type_data['count'], int):
      self.is_fixed = True
      self.count = type_data['count']
    else:
      self.is_foreign = True
      self.count = type_data['count']

    if isinstance(self.field, simple_type):
      self.typename = f"std::vector<{self.field.typename}>"
    else:
      self.typename = f"std::vector<{self.field.name}>"

  def declaration(self):
    if isinstance(self.field, simple_type):
      ret = []
      f_type = self.field.typename
    else:
      self.field.name = f"{self.packet.packet_name}_{self.name}_type"
      ret = self.field.typedef()
      f_type = self.field.name
    if self.is_fixed:
      ret.append(f"std::array<{f_type}, {self.count}> {self.name};")
    else:
      ret.append(f"std::vector<{f_type}> {self.name};")
    return ret

  def fixed(self, encode = True):
    self.field.name = f"el{self.depth}"
    ret = [f"for(auto &el{self.depth} : {self.name}) {{"]
    if encode:
      ret.extend(indent + l for l in self.field.encoder())
    else:
      ret.extend(indent + l for l in self.field.decoder())
    ret.append("}")
    return ret

  def prefixed_encode(self):
    self.count.name = f"{self.name}.size()"
    self.field.name = f"el{self.depth}"
    return [
      *self.count.encoder(),
      f"for(auto &el{self.depth} : {self.name}) {{",
      *(indent + l for l in self.field.encoder()),
      "}"
    ]

  def prefixed_decode(self):
    self.count.name = ''
    self.field.name = f"el{self.depth}"
    return [
      f"{self.name}.resize({self.count.decoder()});",
      f"for(auto &el{self.depth} : {self.name}) {{",
      *(indent + l for l in self.field.decoder()),
      "}"
    ]

  # Identical to switches' compareTo
  def get_foreign(self):
    comp = self.count.replace('../','').replace('..','').replace('/','.')
    p = self.parent
    for i in range(self.count.count('..')):
      while not (isinstance(p, complex_type) or isinstance(p, packet)):
        p = p.parent
      if isinstance(p, packet):
        break;
      else:
        p = p.parent
    while not (isinstance(p, complex_type) or isinstance(p, packet)):
      p = p.parent
    if not isinstance(p, packet):
      comp = f"{p.name}.{comp}"
    return comp

  def foreign_encode(self):
    self.field.name = f"el{self.depth}"
    return [
      f"for(auto &el{self.depth} : {self.name}) {{",
      *(indent + l for l in self.field.encoder()),
      "}"
    ]

  def foreign_decode(self):
    self.field.name = f"el{self.depth}"
    return [
      f"{self.name}.resize({self.get_foreign()});",
      f"for(auto &el{self.depth} : {self.name}) {{",
      *(indent + l for l in self.field.decoder()),
      "}"
    ]

  def encoder(self):
    if self.is_fixed:
      return self.fixed()
    if self.is_prefixed:
      return self.prefixed_encode()
    return self.foreign_encode()

  def decoder(self):
    if self.is_fixed:
      return self.fixed(False)
    if self.is_prefixed:
      return self.prefixed_decode()
    return self.foreign_decode()

# Not a container_type because containers_types sometimes have trivial storage
# requirements. Actual containers always have non-trivial storage, which makes
# them a pure complex type
@mc_data_name('container')
class mc_container(complex_type):
  def __init__(self, name, parent, type_data, use_compare = False):
    super().__init__(name, parent, type_data, use_compare)

    self.fields = []
    for field_info in type_data:
      f_name, f_type, f_data = extract_field(field_info)
      self.fields.append(mcd_typemap[f_type](f_name, self, f_data))

  def encoder(self):
    ret = []
    if self.name:
      for field in self.fields:
        field.temp_name(f"{self.name}.{field.name}")
        ret.extend(field.encoder())
        field.reset_name()
    else:
      for field in self.fields:
        ret.extend(field.encoder())
    return ret

  def decoder(self):
    ret = []
    if self.name:
      for field in self.fields:
        field.temp_name(f"{self.name}.{field.name}")
        ret.extend(field.decoder())
        field.reset_name()
    else:
      for field in self.fields:
        ret.extend(field.decoder())
    return ret

  def __eq__(self, value):
    if not super().__eq__(value) or len(self.fields) != len(value.fields):
        return False
    return all([i == j for i, j in zip(self.fields, value.fields)])


class packet:
  def __init__(self, state, direction, packet_id, packet_name, data):
    self.state = state
    self.direction = "Clientbound" if direction == "toClient" else "Serverbound"
    self.packet_id = packet_id
    self.packet_name = packet_name
    self.data = data
    self.class_name = f"{self.direction}{self.packet_name}"

    self.fields = []
    for data_field in data:
      f_name, f_type, f_data = extract_field(data_field)
      self.fields.append(mcd_typemap[f_type](f_name, self, f_data))

  def declaration(self):
    return [
      f"class {self.class_name} : public Packet {{",
      f"public:",
      *(indent + l for f in self.fields for l in f.declaration()),
      f"{indent}{self.class_name}();",
      f"{indent}void encode(std::ostream &dest) const;",
      f"{indent}void decode(std::istream &src);",
      "};"
    ]

  def constructor(self):
    return [
      f"{self.class_name}::{self.class_name}() :",
      f"{indent*2}Packet({self.state.upper()}, {self.direction.upper()}, "
          f"{self.packet_id},",
      f"{indent*2}\"{self.class_name}\") {{}}"
    ]

  def encoder(self):
    return [
      f"void {self.class_name}::encode(std::ostream &dest) const {{",
      *(indent + l for f in self.fields for l in f.encoder()),
      "}"
    ]

  def decoder(self):
    return [
      f"void {self.class_name}::decode(std::istream &src) {{",
      *(indent + l for f in self.fields for l in f.decoder()),
      "}"
    ]

mc_states = "handshaking", "status", "login", "play"
mc_directions = "toClient", "toServer"

warning = (
  "/*",
  "  This file was generated by mcd2cpp.py",
  "  It should not be edited by hand.",
  "*/",
  "",
)

import re

first_cap_re = re.compile('(.)([A-Z][a-z]+)')
all_cap_re = re.compile('([a-z0-9])([A-Z])')
def to_enum(name, direction, state):
    s1 = first_cap_re.sub(r'\1_\2', name)
    name = all_cap_re.sub(r'\1_\2', s1).upper()
    d = "SB" if direction == "toServer" else "CB"
    st = {"handshaking": "HS", "status": "ST", "login": "LG",
        "play": "PL"}[state]
    return f"{d}_{st}_{name}"

def to_camel_case(string):
  return string.title().replace('_', '')

def extract_infos_from_listing(listing):
  ret = []
  # Why this seemingly random position inside the full listing? Because mcdata
  # hates you.
  for p_id, p_name in listing['types']['packet'][1][0]['type'][1]['mappings'].items():
    ret.append((int(p_id, 0), to_camel_case(p_name), f"packet_{p_name}"))
  return ret

def run(mcd):
  version = mcd.version['minecraftVersion'].replace('.', '_')
  proto = mcd.protocol
  header_upper = [
    *warning,
    f"#ifndef PROTO_{version}_HPP",
    f"#define PROTO_{version}_HPP",
    "",
    "#include <cstdint>",
    "#include <string>",
    "#include <vector>",
    "#include <array>",
    "#include <optional>",
    "#include <memory>",
    "#include <stdexcept>",
    "#include \"datautils.hpp\"",
    "#include \"nbt.hpp\"",
    "",
    f"#define MC_PROTO_VERSION {mcd.version['version']}",
    "",
    "namespace mcd {",
    "",
    "enum packet_direction {",
    "  SERVERBOUND,",
    "  CLIENTBOUND,",
    "  DIRECTION_MAX",
    "};",
    "",
    "enum packet_state {",
    "  HANDSHAKING,",
    "  STATUS,",
    "  LOGIN,",
    "  PLAY,",
    "  STATE_MAX",
    "};",
    ""
  ]
  header_lower = [
    "extern const char* serverbound_handshaking_cstrings["
      "SERVERBOUND_HANDSHAKING_MAX];",
    "extern const char* clientbound_status_cstrings[CLIENTBOUND_STATUS_MAX];",
    "extern const char* serverbound_status_cstrings[SERVERBOUND_STATUS_MAX];",
    "extern const char* clientbound_login_cstrings[CLIENTBOUND_LOGIN_MAX];",
    "extern const char* serverbound_login_cstrings[SERVERBOUND_LOGIN_MAX];",
    "extern const char* clientbound_play_cstrings[CLIENTBOUND_PLAY_MAX];",
    "extern const char* serverbound_play_cstrings[SERVERBOUND_PLAY_MAX];",
    "",
    "extern const char **protocol_cstrings[STATE_MAX][DIRECTION_MAX];",
    "extern const int protocol_max_ids[STATE_MAX][DIRECTION_MAX];",
    "",
    "class Packet {",
    "public:",
    "  const packet_state state;",
    "  const packet_direction direction;",
    "  const std::int32_t packet_id;",
    "  const std::string packet_name;",
    "",
    "  Packet(packet_state state, packet_direction direction,",
    "      std::int32_t packet_id, std::string name) : state(state), ",
    "      direction(direction), packet_id(packet_id), packet_name(name) {}",
    "  virtual ~Packet() = default;",
    "  virtual void encode(std::ostream &buf) const = 0;",
    "  virtual void decode(std::istream &buf) = 0;",
    "};",
    "",
    "std::unique_ptr<Packet> make_packet(packet_state state, "
    "packet_direction dir,",
    f"    int packet_id);",
    ""
  ]
  impl_upper = [
    *warning,
    f"#include \"proto_{version}.hpp\"",
    "",
    "namespace mcd {",
    ""
  ]
  impl_lower = []
  packet_enum = {}
  packets = {}

  for state in mc_states:
    packet_enum[state] = {}
    packets[state] = {}
    for direction in mc_directions:
      packet_enum[state][direction] = []
      packets[state][direction] = []
      packet_infos = extract_infos_from_listing(proto[state][direction])
      for info in packet_infos:
        packet_data = proto[state][direction]['types'][info[2]][1]
        if info[1] != "LegacyServerListPing":
          packet_id = to_enum(info[1], direction, state)
          packet_enum[state][direction].append(packet_id)
        else:
          packet_id = info[0]
        pak = packet(state, direction, packet_id, info[1], packet_data)
        if info[1] != "LegacyServerListPing":
          packets[state][direction].append(pak)
        header_lower += pak.declaration()
        header_lower.append('')

        impl_lower += pak.constructor()
        impl_lower += pak.encoder()
        impl_lower += pak.decoder()
        impl_lower.append('')

  for state in mc_states:
    for direction in mc_directions:
      dr = "clientbound" if direction == "toClient" else "serverbound"
      packet_enum[state][direction].append(f"{dr.upper()}_{state.upper()}_MAX")
      header_upper.append(f"enum {dr}_{state}_ids {{")
      header_upper.extend(f"{indent}{l}," for l in packet_enum[state][direction])
      header_upper[-1] = header_upper[-1][:-1]
      header_upper.extend(("};", ""))

  make_packet = [
    "std::unique_ptr<Packet> make_packet(packet_state state, "
    "packet_direction dir,",
    f"    int packet_id) {{",
    "  switch(state) {"
  ]

  for state in mc_states:
    make_packet.append(f"{indent*2}case {state.upper()}:")
    make_packet.append(f"{indent*3}switch(dir) {{")
    for direction in mc_directions:
      dr = "CLIENTBOUND" if direction == "toClient" else "SERVERBOUND"
      make_packet.append(f"{indent*4}case {dr}:")
      make_packet.append(f"{indent*5}switch(packet_id) {{")
      for pak in packets[state][direction]:
        make_packet.append(f"{indent*6} case {pak.packet_id}:")
        make_packet.append(f"{indent*7} return "
            f"std::make_unique<{pak.class_name}>();")
      make_packet.append(f"{indent*6}default:")
      make_packet.append(f"{indent*7}throw std::runtime_error(\"Invalid "
          "Packet Id\");")
      make_packet.append(f"{indent*5}}}")
    make_packet.append(f"{indent*4}default:")
    make_packet.append(f"{indent*5}throw std::runtime_error(\"Invalid "
        "Packet Direction\");")
    make_packet.append(f"{indent*3}}}")
  make_packet.append(f"{indent*2}default:")
  make_packet.append(f"{indent*3}throw std::runtime_error(\"Invalid "
    "Packet State\");")
  make_packet.extend((f"{indent}}}", "}", ""))

  for state in mc_states:
    for direction in mc_directions:
      if packets[state][direction]:
        dr = "clientbound" if direction == "toClient" else "serverbound"
        impl_upper.append(f"const char* {dr}_{state}_cstrings[] = {{")
        for pak in packets[state][direction]:
          impl_upper.append(f"{indent}\"{pak.class_name}\",")
        impl_upper[-1] = impl_upper[-1][:-1]
        impl_upper.extend(("};", ""))

  impl_upper += [
    "const char **protocol_cstrings[STATE_MAX][DIRECTION_MAX] = {",
    f"{indent}{{serverbound_handshaking_cstrings}},",
    f"{indent}{{serverbound_status_cstrings, clientbound_status_cstrings}},",
    f"{indent}{{serverbound_login_cstrings, clientbound_login_cstrings}},",
    f"{indent}{{serverbound_play_cstrings, clientbound_play_cstrings}}",
    "};",
    "",
    "const int protocol_max_ids[STATE_MAX][DIRECTION_MAX] = {",
    f"{indent}{{SERVERBOUND_HANDSHAKING_MAX, CLIENTBOUND_HANDSHAKING_MAX}},",
    f"{indent}{{SERVERBOUND_STATUS_MAX, CLIENTBOUND_STATUS_MAX}},",
    f"{indent}{{SERVERBOUND_LOGIN_MAX, CLIENTBOUND_LOGIN_MAX}},",
    f"{indent}{{SERVERBOUND_PLAY_MAX, CLIENTBOUND_PLAY_MAX}}",
    "};",
    "",
  ]

  header = header_upper + header_lower + ["}", "#endif", ""]
  impl = impl_upper + impl_lower + make_packet + ["}", ""]


  with open(f"proto_{version}.cpp", "w") as f:
    f.write('\n'.join(impl))
  with open(f"proto_{version}.hpp", "w") as f:
    f.write('\n'.join(header))

if __name__ == "__main__":
  run("1.16.1")



import minecraft_data
from . import blocks, particles, protocol, shapes

def run(version):
  mcd = minecraft_data(version)
  protocol.run(mcd)
  particles.run(mcd)
  shapes.run(mcd)
  blocks.run(mcd)



from rikerbot.PluginBase import PluginBase, pl_announce
from rikerbot import proto


@pl_announce('Start')
class StartPlugin(PluginBase):
  requires = ('Event', 'Exec', 'IO', 'Auth')
  events = {
      'io_connect': 'handle_connect',
  }

  def __init__(self, ploader, settings):
    super().__init__(ploader, settings)
    ploader.provide("Start", self.start)

  def start(self, host="localhost", port="25565"):
    port = str(port)
    self.auth.login()
    self.io.connect(host, port)
    self.exec.run()

  def handle_connect(self, event_id, connect_data):
    packet = proto.ServerboundSetProtocol()
    packet.protocolVersion = proto.MC_PROTO_VERSION
    packet.nextState = proto.LOGIN
    packet.serverHost = connect_data.host
    packet.serverPort = connect_data.port
    self.io.encode_packet(packet)
    packet = proto.ServerboundLoginStart()
    packet.username = self.auth.username
    self.io.encode_packet(packet)



import os
import json

import rikerbot


class SimpleClient:
  def __init__(self, extra_plugins=None, settings=None, online_mode=True, *,
               username='', tokens_path=None):
    extra_plugins = [] if extra_plugins is None else extra_plugins
    settings = {} if settings is None else settings
    if tokens_path is None:
      self.tokens_path = os.path.join(os.getcwd(), "token")

    self.ploader = rikerbot.PluginLoader()
    plugins = rikerbot.default_plugins + extra_plugins
    for failed in rikerbot.solve_dependencies(self.ploader, plugins, settings):
      rikerbot.logger.warning(f"Failed to meet plugin dependency: {failed}")

    ev = self.ploader.require('Event')
    ev.register_callback('auth_login_success', self.write_token)
    self._start = self.ploader.require('Start')
    self.auth = self.ploader.require('Auth')

    self.online_mode = online_mode

    if online_mode:
      if os.path.isfile(self.tokens_path):
        with open(self.tokens_path, 'r') as f:
          tokens = json.loads(f.read())
        self.auth.client_token = tokens['client_token']
        self.auth.access_token = tokens['access_token']
        self.auth.ygg.selected_profile = tokens['selected_profile']
      else:
        self.auth.username = username if username else input("Login Email: ")
        self.auth.password = input("Password: ")
    else:
      self.auth.username = username if username else input("Username: ")

  def __del__(self):
    rikerbot.delete_PluginLoader(self.ploader)

  def start(self, host="localhost", port=25565):
    self._start(host, port)

  def write_token(self, _, __):
    with open(self.tokens_path, 'w') as f:
      f.write(
          json.dumps({
              'client_token': self.auth.client_token,
              'access_token': self.auth.access_token,
              'selected_profile': self.auth.selected_profile
          }))



# This file was automatically generated by SWIG (http://www.swig.org).
# Version 4.0.2
#
# Do not make changes to this file unless you know what you are doing--modify
# the SWIG interface file instead.

from sys import version_info as _swig_python_version_info
if _swig_python_version_info < (2, 7, 0):
    raise RuntimeError("Python 2.7 or later required")

# Import the low-level C/C++ module
if __package__ or "." in __name__:
    from . import _CWorldCore
else:
    import _CWorldCore

try:
    import builtins as __builtin__
except ImportError:
    import __builtin__

def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except __builtin__.Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)


def _swig_setattr_nondynamic_instance_variable(set):
    def set_instance_attr(self, name, value):
        if name == "thisown":
            self.this.own(value)
        elif name == "this":
            set(self, name, value)
        elif hasattr(self, name) and isinstance(getattr(type(self), name), property):
            set(self, name, value)
        else:
            raise AttributeError("You cannot add instance attributes to %s" % self)
    return set_instance_attr


def _swig_setattr_nondynamic_class_variable(set):
    def set_class_attr(cls, name, value):
        if hasattr(cls, name) and not isinstance(getattr(cls, name), property):
            set(cls, name, value)
        else:
            raise AttributeError("You cannot add class attributes to %s" % cls)
    return set_class_attr


def _swig_add_metaclass(metaclass):
    """Class decorator for adding a metaclass to a SWIG wrapped class - a slimmed down version of six.add_metaclass"""
    def wrapper(cls):
        return metaclass(cls.__name__, cls.__bases__, cls.__dict__.copy())
    return wrapper


class _SwigNonDynamicMeta(type):
    """Meta class to enforce nondynamic attributes (no new attributes) for a class"""
    __setattr__ = _swig_setattr_nondynamic_class_variable(type.__setattr__)


class SwigPyIterator(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined - class is abstract")
    __repr__ = _swig_repr
    __swig_destroy__ = _CWorldCore.delete_SwigPyIterator

    def value(self) -> "PyObject *":
        return _CWorldCore.SwigPyIterator_value(self)

    def incr(self, n: "size_t"=1) -> "swig::SwigPyIterator *":
        return _CWorldCore.SwigPyIterator_incr(self, n)

    def decr(self, n: "size_t"=1) -> "swig::SwigPyIterator *":
        return _CWorldCore.SwigPyIterator_decr(self, n)

    def distance(self, x: "SwigPyIterator") -> "ptrdiff_t":
        return _CWorldCore.SwigPyIterator_distance(self, x)

    def equal(self, x: "SwigPyIterator") -> "bool":
        return _CWorldCore.SwigPyIterator_equal(self, x)

    def copy(self) -> "swig::SwigPyIterator *":
        return _CWorldCore.SwigPyIterator_copy(self)

    def next(self) -> "PyObject *":
        return _CWorldCore.SwigPyIterator_next(self)

    def __next__(self) -> "PyObject *":
        return _CWorldCore.SwigPyIterator___next__(self)

    def previous(self) -> "PyObject *":
        return _CWorldCore.SwigPyIterator_previous(self)

    def advance(self, n: "ptrdiff_t") -> "swig::SwigPyIterator *":
        return _CWorldCore.SwigPyIterator_advance(self, n)

    def __eq__(self, x: "SwigPyIterator") -> "bool":
        return _CWorldCore.SwigPyIterator___eq__(self, x)

    def __ne__(self, x: "SwigPyIterator") -> "bool":
        return _CWorldCore.SwigPyIterator___ne__(self, x)

    def __iadd__(self, n: "ptrdiff_t") -> "swig::SwigPyIterator &":
        return _CWorldCore.SwigPyIterator___iadd__(self, n)

    def __isub__(self, n: "ptrdiff_t") -> "swig::SwigPyIterator &":
        return _CWorldCore.SwigPyIterator___isub__(self, n)

    def __add__(self, n: "ptrdiff_t") -> "swig::SwigPyIterator *":
        return _CWorldCore.SwigPyIterator___add__(self, n)

    def __sub__(self, *args) -> "ptrdiff_t":
        return _CWorldCore.SwigPyIterator___sub__(self, *args)
    def __iter__(self):
        return self

# Register SwigPyIterator in _CWorldCore:
_CWorldCore.SwigPyIterator_swigregister(SwigPyIterator)

class BlockCoord(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    x = property(_CWorldCore.BlockCoord_x_get, _CWorldCore.BlockCoord_x_set)
    y = property(_CWorldCore.BlockCoord_y_get, _CWorldCore.BlockCoord_y_set)
    z = property(_CWorldCore.BlockCoord_z_get, _CWorldCore.BlockCoord_z_set)

    def __init__(self):
        _CWorldCore.BlockCoord_swiginit(self, _CWorldCore.new_BlockCoord())
    __swig_destroy__ = _CWorldCore.delete_BlockCoord

# Register BlockCoord in _CWorldCore:
_CWorldCore.BlockCoord_swigregister(BlockCoord)

class BlockIdVector(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def iterator(self) -> "swig::SwigPyIterator *":
        return _CWorldCore.BlockIdVector_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __nonzero__(self) -> "bool":
        return _CWorldCore.BlockIdVector___nonzero__(self)

    def __bool__(self) -> "bool":
        return _CWorldCore.BlockIdVector___bool__(self)

    def __len__(self) -> "std::vector< unsigned short >::size_type":
        return _CWorldCore.BlockIdVector___len__(self)

    def __getslice__(self, i: "std::vector< unsigned short >::difference_type", j: "std::vector< unsigned short >::difference_type") -> "std::vector< unsigned short,std::allocator< unsigned short > > *":
        return _CWorldCore.BlockIdVector___getslice__(self, i, j)

    def __setslice__(self, *args) -> "void":
        return _CWorldCore.BlockIdVector___setslice__(self, *args)

    def __delslice__(self, i: "std::vector< unsigned short >::difference_type", j: "std::vector< unsigned short >::difference_type") -> "void":
        return _CWorldCore.BlockIdVector___delslice__(self, i, j)

    def __delitem__(self, *args) -> "void":
        return _CWorldCore.BlockIdVector___delitem__(self, *args)

    def __getitem__(self, *args) -> "std::vector< unsigned short >::value_type const &":
        return _CWorldCore.BlockIdVector___getitem__(self, *args)

    def __setitem__(self, *args) -> "void":
        return _CWorldCore.BlockIdVector___setitem__(self, *args)

    def pop(self) -> "std::vector< unsigned short >::value_type":
        return _CWorldCore.BlockIdVector_pop(self)

    def append(self, x: "std::vector< unsigned short >::value_type const &") -> "void":
        return _CWorldCore.BlockIdVector_append(self, x)

    def empty(self) -> "bool":
        return _CWorldCore.BlockIdVector_empty(self)

    def size(self) -> "std::vector< unsigned short >::size_type":
        return _CWorldCore.BlockIdVector_size(self)

    def swap(self, v: "BlockIdVector") -> "void":
        return _CWorldCore.BlockIdVector_swap(self, v)

    def begin(self) -> "std::vector< unsigned short >::iterator":
        return _CWorldCore.BlockIdVector_begin(self)

    def end(self) -> "std::vector< unsigned short >::iterator":
        return _CWorldCore.BlockIdVector_end(self)

    def rbegin(self) -> "std::vector< unsigned short >::reverse_iterator":
        return _CWorldCore.BlockIdVector_rbegin(self)

    def rend(self) -> "std::vector< unsigned short >::reverse_iterator":
        return _CWorldCore.BlockIdVector_rend(self)

    def clear(self) -> "void":
        return _CWorldCore.BlockIdVector_clear(self)

    def get_allocator(self) -> "std::vector< unsigned short >::allocator_type":
        return _CWorldCore.BlockIdVector_get_allocator(self)

    def pop_back(self) -> "void":
        return _CWorldCore.BlockIdVector_pop_back(self)

    def erase(self, *args) -> "std::vector< unsigned short >::iterator":
        return _CWorldCore.BlockIdVector_erase(self, *args)

    def __init__(self, *args):
        _CWorldCore.BlockIdVector_swiginit(self, _CWorldCore.new_BlockIdVector(*args))

    def push_back(self, x: "std::vector< unsigned short >::value_type const &") -> "void":
        return _CWorldCore.BlockIdVector_push_back(self, x)

    def front(self) -> "std::vector< unsigned short >::value_type const &":
        return _CWorldCore.BlockIdVector_front(self)

    def back(self) -> "std::vector< unsigned short >::value_type const &":
        return _CWorldCore.BlockIdVector_back(self)

    def assign(self, n: "std::vector< unsigned short >::size_type", x: "std::vector< unsigned short >::value_type const &") -> "void":
        return _CWorldCore.BlockIdVector_assign(self, n, x)

    def resize(self, *args) -> "void":
        return _CWorldCore.BlockIdVector_resize(self, *args)

    def insert(self, *args) -> "void":
        return _CWorldCore.BlockIdVector_insert(self, *args)

    def reserve(self, n: "std::vector< unsigned short >::size_type") -> "void":
        return _CWorldCore.BlockIdVector_reserve(self, n)

    def capacity(self) -> "std::vector< unsigned short >::size_type":
        return _CWorldCore.BlockIdVector_capacity(self)
    __swig_destroy__ = _CWorldCore.delete_BlockIdVector

# Register BlockIdVector in _CWorldCore:
_CWorldCore.BlockIdVector_swigregister(BlockIdVector)

class PositionVector(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def iterator(self) -> "swig::SwigPyIterator *":
        return _CWorldCore.PositionVector_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __nonzero__(self) -> "bool":
        return _CWorldCore.PositionVector___nonzero__(self)

    def __bool__(self) -> "bool":
        return _CWorldCore.PositionVector___bool__(self)

    def __len__(self) -> "std::vector< mcd::mc_position >::size_type":
        return _CWorldCore.PositionVector___len__(self)

    def __getslice__(self, i: "std::vector< mcd::mc_position >::difference_type", j: "std::vector< mcd::mc_position >::difference_type") -> "std::vector< mcd::mc_position,std::allocator< mcd::mc_position > > *":
        return _CWorldCore.PositionVector___getslice__(self, i, j)

    def __setslice__(self, *args) -> "void":
        return _CWorldCore.PositionVector___setslice__(self, *args)

    def __delslice__(self, i: "std::vector< mcd::mc_position >::difference_type", j: "std::vector< mcd::mc_position >::difference_type") -> "void":
        return _CWorldCore.PositionVector___delslice__(self, i, j)

    def __delitem__(self, *args) -> "void":
        return _CWorldCore.PositionVector___delitem__(self, *args)

    def __getitem__(self, *args) -> "std::vector< mcd::mc_position >::value_type const &":
        return _CWorldCore.PositionVector___getitem__(self, *args)

    def __setitem__(self, *args) -> "void":
        return _CWorldCore.PositionVector___setitem__(self, *args)

    def pop(self) -> "std::vector< mcd::mc_position >::value_type":
        return _CWorldCore.PositionVector_pop(self)

    def append(self, x: "std::vector< mcd::mc_position >::value_type const &") -> "void":
        return _CWorldCore.PositionVector_append(self, x)

    def empty(self) -> "bool":
        return _CWorldCore.PositionVector_empty(self)

    def size(self) -> "std::vector< mcd::mc_position >::size_type":
        return _CWorldCore.PositionVector_size(self)

    def swap(self, v: "PositionVector") -> "void":
        return _CWorldCore.PositionVector_swap(self, v)

    def begin(self) -> "std::vector< mcd::mc_position >::iterator":
        return _CWorldCore.PositionVector_begin(self)

    def end(self) -> "std::vector< mcd::mc_position >::iterator":
        return _CWorldCore.PositionVector_end(self)

    def rbegin(self) -> "std::vector< mcd::mc_position >::reverse_iterator":
        return _CWorldCore.PositionVector_rbegin(self)

    def rend(self) -> "std::vector< mcd::mc_position >::reverse_iterator":
        return _CWorldCore.PositionVector_rend(self)

    def clear(self) -> "void":
        return _CWorldCore.PositionVector_clear(self)

    def get_allocator(self) -> "std::vector< mcd::mc_position >::allocator_type":
        return _CWorldCore.PositionVector_get_allocator(self)

    def pop_back(self) -> "void":
        return _CWorldCore.PositionVector_pop_back(self)

    def erase(self, *args) -> "std::vector< mcd::mc_position >::iterator":
        return _CWorldCore.PositionVector_erase(self, *args)

    def __init__(self, *args):
        _CWorldCore.PositionVector_swiginit(self, _CWorldCore.new_PositionVector(*args))

    def push_back(self, x: "std::vector< mcd::mc_position >::value_type const &") -> "void":
        return _CWorldCore.PositionVector_push_back(self, x)

    def front(self) -> "std::vector< mcd::mc_position >::value_type const &":
        return _CWorldCore.PositionVector_front(self)

    def back(self) -> "std::vector< mcd::mc_position >::value_type const &":
        return _CWorldCore.PositionVector_back(self)

    def assign(self, n: "std::vector< mcd::mc_position >::size_type", x: "std::vector< mcd::mc_position >::value_type const &") -> "void":
        return _CWorldCore.PositionVector_assign(self, n, x)

    def resize(self, *args) -> "void":
        return _CWorldCore.PositionVector_resize(self, *args)

    def insert(self, *args) -> "void":
        return _CWorldCore.PositionVector_insert(self, *args)

    def reserve(self, n: "std::vector< mcd::mc_position >::size_type") -> "void":
        return _CWorldCore.PositionVector_reserve(self, n)

    def capacity(self) -> "std::vector< mcd::mc_position >::size_type":
        return _CWorldCore.PositionVector_capacity(self)
    __swig_destroy__ = _CWorldCore.delete_PositionVector

# Register PositionVector in _CWorldCore:
_CWorldCore.PositionVector_swigregister(PositionVector)

class PosArray(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def iterator(self) -> "swig::SwigPyIterator *":
        return _CWorldCore.PosArray_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __nonzero__(self) -> "bool":
        return _CWorldCore.PosArray___nonzero__(self)

    def __bool__(self) -> "bool":
        return _CWorldCore.PosArray___bool__(self)

    def __len__(self) -> "std::array< int,3 >::size_type":
        return _CWorldCore.PosArray___len__(self)

    def __getslice__(self, i: "std::array< int,3 >::difference_type", j: "std::array< int,3 >::difference_type") -> "std::array< std::int32_t,3 > *":
        return _CWorldCore.PosArray___getslice__(self, i, j)

    def __setslice__(self, *args) -> "void":
        return _CWorldCore.PosArray___setslice__(self, *args)

    def __delslice__(self, i: "std::array< int,3 >::difference_type", j: "std::array< int,3 >::difference_type") -> "void":
        return _CWorldCore.PosArray___delslice__(self, i, j)

    def __delitem__(self, *args) -> "void":
        return _CWorldCore.PosArray___delitem__(self, *args)

    def __getitem__(self, *args) -> "std::array< int,3 >::value_type const &":
        return _CWorldCore.PosArray___getitem__(self, *args)

    def __setitem__(self, *args) -> "void":
        return _CWorldCore.PosArray___setitem__(self, *args)

    def __init__(self, *args):
        _CWorldCore.PosArray_swiginit(self, _CWorldCore.new_PosArray(*args))

    def empty(self) -> "bool":
        return _CWorldCore.PosArray_empty(self)

    def size(self) -> "std::array< int,3 >::size_type":
        return _CWorldCore.PosArray_size(self)

    def swap(self, v: "PosArray") -> "void":
        return _CWorldCore.PosArray_swap(self, v)

    def begin(self) -> "std::array< int,3 >::iterator":
        return _CWorldCore.PosArray_begin(self)

    def end(self) -> "std::array< int,3 >::iterator":
        return _CWorldCore.PosArray_end(self)

    def rbegin(self) -> "std::array< int,3 >::reverse_iterator":
        return _CWorldCore.PosArray_rbegin(self)

    def rend(self) -> "std::array< int,3 >::reverse_iterator":
        return _CWorldCore.PosArray_rend(self)

    def front(self) -> "std::array< int,3 >::value_type const &":
        return _CWorldCore.PosArray_front(self)

    def back(self) -> "std::array< int,3 >::value_type const &":
        return _CWorldCore.PosArray_back(self)

    def fill(self, u: "std::array< int,3 >::value_type const &") -> "void":
        return _CWorldCore.PosArray_fill(self, u)
    __swig_destroy__ = _CWorldCore.delete_PosArray

# Register PosArray in _CWorldCore:
_CWorldCore.PosArray_swigregister(PosArray)

class PosArrayVector(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def iterator(self) -> "swig::SwigPyIterator *":
        return _CWorldCore.PosArrayVector_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __nonzero__(self) -> "bool":
        return _CWorldCore.PosArrayVector___nonzero__(self)

    def __bool__(self) -> "bool":
        return _CWorldCore.PosArrayVector___bool__(self)

    def __len__(self) -> "std::vector< std::array< int,3 > >::size_type":
        return _CWorldCore.PosArrayVector___len__(self)

    def __getslice__(self, i: "std::vector< std::array< int,3 > >::difference_type", j: "std::vector< std::array< int,3 > >::difference_type") -> "std::vector< std::array< std::int32_t,3 >,std::allocator< std::array< std::int32_t,3 > > > *":
        return _CWorldCore.PosArrayVector___getslice__(self, i, j)

    def __setslice__(self, *args) -> "void":
        return _CWorldCore.PosArrayVector___setslice__(self, *args)

    def __delslice__(self, i: "std::vector< std::array< int,3 > >::difference_type", j: "std::vector< std::array< int,3 > >::difference_type") -> "void":
        return _CWorldCore.PosArrayVector___delslice__(self, i, j)

    def __delitem__(self, *args) -> "void":
        return _CWorldCore.PosArrayVector___delitem__(self, *args)

    def __getitem__(self, *args) -> "std::vector< std::array< int,3 > >::value_type const &":
        return _CWorldCore.PosArrayVector___getitem__(self, *args)

    def __setitem__(self, *args) -> "void":
        return _CWorldCore.PosArrayVector___setitem__(self, *args)

    def pop(self) -> "std::vector< std::array< int,3 > >::value_type":
        return _CWorldCore.PosArrayVector_pop(self)

    def append(self, x: "PosArray") -> "void":
        return _CWorldCore.PosArrayVector_append(self, x)

    def empty(self) -> "bool":
        return _CWorldCore.PosArrayVector_empty(self)

    def size(self) -> "std::vector< std::array< int,3 > >::size_type":
        return _CWorldCore.PosArrayVector_size(self)

    def swap(self, v: "PosArrayVector") -> "void":
        return _CWorldCore.PosArrayVector_swap(self, v)

    def begin(self) -> "std::vector< std::array< int,3 > >::iterator":
        return _CWorldCore.PosArrayVector_begin(self)

    def end(self) -> "std::vector< std::array< int,3 > >::iterator":
        return _CWorldCore.PosArrayVector_end(self)

    def rbegin(self) -> "std::vector< std::array< int,3 > >::reverse_iterator":
        return _CWorldCore.PosArrayVector_rbegin(self)

    def rend(self) -> "std::vector< std::array< int,3 > >::reverse_iterator":
        return _CWorldCore.PosArrayVector_rend(self)

    def clear(self) -> "void":
        return _CWorldCore.PosArrayVector_clear(self)

    def get_allocator(self) -> "std::vector< std::array< int,3 > >::allocator_type":
        return _CWorldCore.PosArrayVector_get_allocator(self)

    def pop_back(self) -> "void":
        return _CWorldCore.PosArrayVector_pop_back(self)

    def erase(self, *args) -> "std::vector< std::array< int,3 > >::iterator":
        return _CWorldCore.PosArrayVector_erase(self, *args)

    def __init__(self, *args):
        _CWorldCore.PosArrayVector_swiginit(self, _CWorldCore.new_PosArrayVector(*args))

    def push_back(self, x: "PosArray") -> "void":
        return _CWorldCore.PosArrayVector_push_back(self, x)

    def front(self) -> "std::vector< std::array< int,3 > >::value_type const &":
        return _CWorldCore.PosArrayVector_front(self)

    def back(self) -> "std::vector< std::array< int,3 > >::value_type const &":
        return _CWorldCore.PosArrayVector_back(self)

    def assign(self, n: "std::vector< std::array< int,3 > >::size_type", x: "PosArray") -> "void":
        return _CWorldCore.PosArrayVector_assign(self, n, x)

    def resize(self, *args) -> "void":
        return _CWorldCore.PosArrayVector_resize(self, *args)

    def insert(self, *args) -> "void":
        return _CWorldCore.PosArrayVector_insert(self, *args)

    def reserve(self, n: "std::vector< std::array< int,3 > >::size_type") -> "void":
        return _CWorldCore.PosArrayVector_reserve(self, n)

    def capacity(self) -> "std::vector< std::array< int,3 > >::size_type":
        return _CWorldCore.PosArrayVector_capacity(self)
    __swig_destroy__ = _CWorldCore.delete_PosArrayVector

# Register PosArrayVector in _CWorldCore:
_CWorldCore.PosArrayVector_swigregister(PosArrayVector)

class PluginBase(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    type_query = property(_CWorldCore.PluginBase_type_query_get)

    def __init__(self, *args):
        _CWorldCore.PluginBase_swiginit(self, _CWorldCore.new_PluginBase(*args))
    __swig_destroy__ = _CWorldCore.delete_PluginBase

# Register PluginBase in _CWorldCore:
_CWorldCore.PluginBase_swigregister(PluginBase)

class WorldCore(PluginBase):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def __init__(self, ploader: "rkr::PluginLoader &", ownership: "bool"=False):
        _CWorldCore.WorldCore_swiginit(self, _CWorldCore.new_WorldCore(ploader, ownership))

    def get(self, *args) -> "std::vector< rkr::block_id,std::allocator< rkr::block_id > >":
        return _CWorldCore.WorldCore_get(self, *args)
    __swig_destroy__ = _CWorldCore.delete_WorldCore

# Register WorldCore in _CWorldCore:
_CWorldCore.WorldCore_swigregister(WorldCore)






import copy


# "plugins" is a list of (name, plugin) tuples, where "name" corresponds to an
# entry in the settings dict
def solve_dependencies(ploader, plugins, settings=None):
  announce = {}
  loaded = []
  failed_dependencies = []
  plugins = copy.copy(plugins)
  settings = {} if settings is None else settings

  for name, plugin in plugins:
    if hasattr(plugin, 'pl_announce'):
      for ident in plugin.pl_announce:
        announce[ident] = (name, plugin)

  def load_plugin(plugin_tuple):
    name, plugin = plugin_tuple
    requirements = plugin.requires
    if isinstance(requirements, str):
      requirements = (requirements, )
    for requirement in requirements:
      if requirement in announce and requirement not in loaded:
        load_plugin(announce[requirement])
      elif requirement not in announce:
        failed_dependencies.append(requirement)
    plugin(ploader, settings.get(name, {}))
    if hasattr(plugin, 'pl_announce'):
      loaded.extend(plugin.pl_announce)
    plugins.remove(plugin_tuple)

  # Event is loaded implicitly
  load_plugin(announce['Event'])

  while plugins:
    load_plugin(plugins[0])

  return failed_dependencies



import sys

version = sys.argv[1].replace(".", "_")
swig_interface = (
  f'%module Proto{version}',
  '%{',
  f'#include "proto_{version}.hpp"',
  '%}',
  '%feature ("flatnested");',
  '',
  '%include <stdint.i>',
  '%include <std_string.i>',
  '%include "typemaps/unique_ptr.i"',
  '%unique_ptr(mcd::Packet)',
  '%rename(_property) property;',
  '',
  '%typemap(out) (std::vector<char>) {',
  '  $result = PyBytes_FromStringAndSize($1.data(), $1.size());',
  '}',
  '',
  '%include "datautils.hpp"',
  '%include "particletypes.hpp"',
  f'%include "proto_{version}.hpp"',
  '',
)

with open(f"Proto{version}.i", "w") as f:
  f.write("\n".join(swig_interface))

with open(f"MinecraftProtocol.py", "w") as f:
  f.write(f"from .Proto{version} import *\n")

with open(f"minecraft_protocol.hpp", "w") as f:
  f.write(f'#include "proto_{version}.hpp"\n')

import mcd2cpp
mcd2cpp.run(sys.argv[1])



from rikerbot.PluginBase import CPluginBase, pl_announce
from .CIOCore import IOCore


@pl_announce('IO')
class IOPlugin(CPluginBase):
  requires = ('Event', 'Exec')
  core = IOCore

  def __init__(self, ploader, settings) -> None:
    self.core(ploader, ploader.require('Exec').get_ctx(), True).thisown = 0



import json
from uuid import uuid4
from urllib.request import urlopen, Request
from urllib.error import HTTPError


class Yggdrasil(object):
  ygg_version = 1
  ygg_url = 'https://authserver.mojang.com'

  def __init__(self, username='', password='', client_token='',
               access_token=''):
    self.username = username
    self.password = password
    self.client_token = client_token
    self.access_token = access_token
    self.available_profiles = []
    self.selected_profile = {}

  def login(self):
    if self.access_token and self.validate():
      return True
    if self.access_token and self.client_token and self.refresh():
      return True
    self.client_token = uuid4().hex
    return self.username and self.password and self.authenticate()

  def logout(self):
    return self.access_token and self.client_token and self.invalidate()

  def _ygg_req(self, endpoint, payload):
    try:
      resp = urlopen(
          Request(url=self.ygg_url + endpoint,
                  data=json.dumps(payload).encode('utf-8'),
                  headers={'Content-Type': 'application/json'}))
    except HTTPError as e:
      resp = e
    data = resp.read().decode('utf-8')
    return json.loads(data) if data else dict()

  def authenticate(self):
    """
    Generate an access token using an username and password. Any existing
    client token is invalidated if not provided.
    Returns:
      dict: Response or error dict
    """
    endpoint = '/authenticate'

    payload = {
        'agent': {
            'name': 'Minecraft',
            'version': self.ygg_version,
        },
        'username': self.username,
        'password': self.password,
        'clientToken': self.client_token,
    }
    rep = self._ygg_req(endpoint, payload)
    if not rep or 'error' in rep:
      return False
    self.access_token = rep['accessToken']
    self.client_token = rep['clientToken']
    self.available_profiles = rep['availableProfiles']
    self.selected_profile = rep['selectedProfile']
    return True

  def refresh(self):
    """
    Generate an access token with a client/access token pair. Used
    access token is invalidated.
    Returns:
      dict: Response or error dict
    """
    endpoint = '/refresh'

    payload = {
        'accessToken': self.access_token,
        'clientToken': self.client_token,
    }
    rep = self._ygg_req(endpoint, payload)
    if not rep or 'error' in rep:
      return False

    self.access_token = rep['accessToken']
    self.client_token = rep['clientToken']
    self.selected_profile = rep['selectedProfile']
    return True

  def signout(self):
    """
    Invalidate access tokens with a username and password.
    Returns:
      dict: Empty or error dict
    """
    endpoint = '/signout'

    payload = {
        'username': self.username,
        'password': self.password,
    }
    rep = self._ygg_req(endpoint, payload)
    if not rep or 'error' in rep:
      return False
    self.client_token = ''
    self.access_token = ''
    self.available_profiles = []
    self.selected_profile = {}
    return True

  def invalidate(self):
    """
    Invalidate access tokens with a client/access token pair
    Returns:
      dict: Empty or error dict
    """
    endpoint = '/invalidate'

    payload = {
        'accessToken': self.access_token,
        'clientToken': self.client_token,
    }
    self._ygg_req(endpoint, payload)
    self.client_token = ''
    self.access_token = ''
    self.available_profiles = []
    self.selected_profile = {}
    return True

  def validate(self):
    """
    Check if an access token is valid
    Returns:
      dict: Empty or error dict
    """
    endpoint = '/validate'

    payload = dict(accessToken=self.access_token)
    rep = self._ygg_req(endpoint, payload)
    return not bool(rep)



def run(mcd, indent = '  '):
  i = indent
  prtcle_hdr = [
    "/*",
    f"{i}This file was generated by mcd2cpp.py",
    f"{i}It should not be edited by hand.",
    "*/",
    "",
    f"// For MCD version {mcd.version['minecraftVersion']}",
    "#ifndef PARTICLETYPES_HPP",
    "#define PARTICLETYPES_HPP",
    "",
    "namespace mcd {",
    "",
    "enum particle_type {"
  ]
  prtcle_hdr.extend(f"{indent}PARTICLE_{x.upper()}," for x in
      mcd.particles_name)
  prtcle_hdr[-1] = prtcle_hdr[-1][:-1]
  prtcle_hdr += ["};", "}", "#endif", ""]

  with open(f"particletypes.hpp", "w") as f:
    f.write('\n'.join(prtcle_hdr))



# forget the ink, the milk, the blood—
# all was washed clean with the flood
# we rose up from the falling waters
# the fallen rain’s own sons and daughters
#
# and none of this, none of this matters.
#   - Don Paterson, Rain

from rikerbot.PluginBase import PluginBase
from rikerbot import proto


class KeepAlivePlugin(PluginBase):
  requires = ('IO')
  events = {'ClientboundKeepAlive': 'handle_keep_alive'}

  def __init__(self, ploader, settings):
    super().__init__(ploader, settings)

  def handle_keep_alive(self, event_id, in_packet):
    out_packet = proto.ServerboundKeepAlive()
    out_packet.keepAliveId = in_packet.keepAliveId
    self.io.encode_packet(out_packet)



from setuptools import setup, find_packages, Extension, Distribution
from setuptools.command.build_ext import build_ext
import os
import pathlib
import shutil

suffix = '.pyd' if os.name == 'nt' else '.so'
build_tool = 'Ninja' if shutil.which('ninja') else 'Unix Makefiles'
build_type = 'Release'


class RKRDistribution(Distribution):
  def iter_distribution_names(self):
    for pkg in self.packages or ():
      yield pkg
    for module in self.py_modules or ():
      yield module


class RKRExtension(Extension):
  def __init__(self, path):
    self.path = path
    super().__init__(pathlib.PurePath(path).name, [])


def find_extensions(directory):
  extensions = []
  for path, _, filenames in os.walk(directory):
    for filename in filenames:
      filename = pathlib.PurePath(filename)
      if pathlib.PurePath(filename).suffix == suffix:
        extensions.append(RKRExtension(os.path.join(path, filename.stem)))
  return extensions


class build_RKRExtensions(build_ext):
  def run(self):
    self.announce("Configuring CMake", level=3)
    source_dir = pathlib.PurePath(__file__).parent
    build_dir = source_dir / 'build' / build_type

    self.spawn([
        'cmake', '--no-warn-unused-cli',
        '-DCMAKE_EXPORT_COMPILE_COMMANDS:BOOL=TRUE',
        f'-DCMAKE_BUILD_TYPE:STRING={build_type}', f'-G{build_tool}',
        f'-H{source_dir}', f'-B{build_dir}'
    ])

    self.announce("Building binaries", level=3)

    self.spawn([
        'cmake', '--build',
        str(build_dir), '--config',
        str(build_type), '--target', 'rikerbot_all', '-j', '14'
    ])

    self.extensions = find_extensions('rikerbot')

    for ext in self.extensions:
      source = f"{ext.path}{suffix}"
      ext_dir = pathlib.PurePath(self.get_ext_fullpath(ext.name)).parent
      os.makedirs(f"{ext_dir / pathlib.PurePath(ext.path).parent}",
                  exist_ok=True)
      shutil.copy(f"{source}", f"{ext_dir}/{source}")


setup(
    name='rikerbot',  # This comment prevents yapf formatting
    description='RikerBot is a framework for creating Python Minecraft Bots '
    'with C++ extensions',
    license='zlib',
    long_description=open('ReadMe.rst').read(),
    version='0.0.2',
    url='https://github.com/SpockBotMC/RikerBot',
    packages=find_packages(exclude=["mcd2cpp"]),
    keywords=['minecraft'],
    author="N. Vito Gamberini",
    author_email="vito@gamberini.email",
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: zlib/libpng License',
        'Programming Language :: C++',
        'Programming Language :: Python :: 3 :: Only'
    ],
    ext_modules=[RKRExtension("rikerbot")],
    distclass=RKRDistribution,
    cmdclass={'build_ext': build_RKRExtensions})



# Okay so here's the deal:
# Ownership of objects that get passed between C++ and Python is tricky.
# Generally speaking, the best idea is for C++/Python to share ownership of
# pure Python objects, and for C++ to take ownership of pure C++ ones.
#
# In practice this is handled by the Plugin Loader, which will take over
# ownership of C++ objects and stores them in an "owned" vector, and for Python
# objects it will correctly increment/decrement their reference count. The
# problem then is this, quis custodiet ipsos custodes? Who owns the plugin
# loaders?
plugin_loader_list = []
# ^ This list right here does, take that Socrates
#
# You may wonder why this is necessary. It's necessary because if you build any
# sort of "Client" class you're going to be tempted to drop the reference
# you're holding to the PluginLoader, it's not needed so why keep it? Because
# it owns everything, and if you drop it you will segfault and not understand
# why. And then you will file a bug report, and I will have to explain that
# Python lied to you and memory management does matter.
#
# We could of course just leak the PluginLoaders' memory, allow them to ride
# eternal, shiny and chrome! Realistically they're going to be around for the
# life of the program anyway, and there is rarely more than one. But this is
# sacrilege in all of the popular programming sects, so we don't do that.
#
# Thus my list and I, we own the plugin loader, so you don't have to.
# You're welcome.

from .CPluginLoader import PluginLoader as _PluginLoader


class PluginLoader(_PluginLoader):
  def __init__(self):
    super().__init__()
    if plugin_loader_list:
      self.pl_id = plugin_loader_list[-1].pl_id + 1
    else:
      self.pl_id = 0
    plugin_loader_list.append(self)


def make_PluginLoader():
  pl = PluginLoader()
  return pl.pl_id, pl


def delete_PluginLoader(pl):
  if isinstance(pl, int):
    pl = next(filter(lambda x: x.pl_id == pl, plugin_loader_list), None)
  if not pl or pl not in plugin_loader_list:
    return None
  plugin_loader_list.remove(pl)
  return pl.pl_id



def run(mcd, indent = '  '):
  i = indent
  block_hdr = [
    "/*",
    f"{i}This file was generated by mcd2cpp.py",
    f"{i}It should not be edited by hand.",
    "*/",
    "",
    f"// For MCD version {mcd.version['minecraftVersion']}",
    "",
    "#ifndef BLOCKDATA_HPP",
    "#define BLOCKDATA_HPP",
    "",
    "#include <cstdint>",
    "#include <string>",
    "#include <vector>",
    "#include <optional>",
    "#include <functional>",
    "#include \"shape_data.hpp\"",
    "",
    "namespace mcd {",
    "",
    "struct MCBlockPropertyData {",
    f"{i}std::string name;",
    f"{i}std::uint8_t num_values;",
    "};",
    "",
    "struct MCBlockData {",
    f"{i}std::uint64_t id;",
    f"{i}std::uint64_t min_state_id;",
    f"{i}std::uint64_t max_state_id;",
    f"{i}std::uint64_t default_state;",
    f"{i}std::uint16_t stack_size;",
    f"{i}std::uint8_t emit_light;",
    f"{i}std::uint8_t filter_light;",
    "",
    f"{i}std::string name;",
    f"{i}std::string display_name;",
    "",
    f"{i}bool diggable;",
    f"{i}bool collidable;",
    f"{i}bool transparent;",
    "",
    f"{i}std::vector<std::reference_wrapper<const std::vector<MCBBoxData>>> shapes;",
    f"{i}std::vector<std::uint64_t> drops;",
    f"{i}std::vector<MCBlockPropertyData> props;",
    "",
    f"{i}std::optional<std::string> material;",
    f"{i}std::optional<float> hardness;",
    "};",
    "",
    f"extern const MCBlockData block_data[{len(mcd.blocks_list)}];",
    "",
    "const MCBlockData& from_stateID(std::uint64_t state_id);"
    "",
    "} // namespace mcd",
    "#endif // BLOCKDATA_HPP",
  ]

  with open(f"block_data.hpp", "w") as f:
    f.write('\n'.join(block_hdr))

  block_cpp = [
    "/*",
    f"{i}This file was generated by mcd2cpp.py",
    f"{i}It should not be edited by hand.",
    "*/",
    "",
    "#include <map>",
    "#include \"block_data.hpp\"",
    "",
    "namespace mcd {",
    ""
    "const MCBlockData block_data[] = {",
  ]

  for block in mcd.blocks_list:
    shapes = mcd.blockCollisionShapes['blocks'][block['name']]
    if isinstance(shapes, int):
      shapes = [shapes]
    block_cpp.extend((
      f"{i}{{",
      f"{i*2}.id = {block['id']},",
      f"{i*2}.min_state_id = {block['minStateId']},",
      f"{i*2}.max_state_id = {block['maxStateId']},",
      f"{i*2}.default_state = {block['defaultState']},",
      f"{i*2}.stack_size = {block['stackSize']},",
      f"{i*2}.emit_light = {block['emitLight']},",
      f"{i*2}.filter_light = {block['filterLight']},",
      "",
      f"{i*2}.name = \"{block['name']}\",",
      f"{i*2}.display_name = \"{block['displayName']}\",",
      "",
      f"{i*2}.diggable = {str(block['diggable']).lower()},",
      f"{i*2}.collidable = {str(block['boundingBox'] == 'block').lower()},",
      f"{i*2}.transparent = {str(block['transparent']).lower()},",
      "",
      f"{i*2}.shapes = {{{', '.join(f'shapes[{x}]' for x in shapes)}}},",
      f"{i*2}.drops = {{{', '.join(str(x) for x in block['drops'])}}},",
    ))

    if block['states']:
      block_cpp.append(f"{i*2}.props = {{")
      block_cpp.extend(f"{i*3}{{\"{prop['name']}\", {prop['num_values']}}},"
          for prop in block['states'])
      block_cpp[-1] = block_cpp[-1][:-1]
      block_cpp.append(f"{i*2}}}")
    else:
      block_cpp.append(f"{i*2}.props = {{}}")

    if 'material' in block:
      block_cpp[-1] += ','
      block_cpp.append(f"{i*2}.material = \"{block['material']}\"")

    if block['hardness'] is not None:
      block_cpp[-1] += ','
      block_cpp.append(f"{i*2}.hardness = {block['hardness']}")

    block_cpp.append(f"{i}}},")

  block_cpp[-1] = block_cpp[-1][:-1]
  block_cpp.extend((
    "};",
    "",
    "const std::map<std::uint64_t, const MCBlockData&> state_id_map = {"
  ))

  block_cpp.append(',\n'.join(f"{i}{{{block['maxStateId']}, block_data[{block['id']}]}}" for block in mcd.blocks_list))

  block_cpp.extend((
    "};",
    "",
    "const MCBlockData& from_stateID(std::uint64_t state_id) {",
    f"{i}return state_id_map.lower_bound(state_id)->second;",
    "}",
    "",
    "} // namespace mcd",
    ""
  ))

  with open(f"block_data.cpp", "w") as f:
    f.write('\n'.join(block_cpp))
